"""
Eigenmode analysis (EMA)
-------------------------------------------------------------

This method, if needed, calculates modes (they can also be read in from a file)
and perturbs the spin system periodically in the direction of the eigenmode.
"""

import spirit.spiritlib as spiritlib
import ctypes

### Load Library
_spirit = spiritlib.load_spirit_library()

## ---------------------------------- Set ----------------------------------

_EMA_Set_N_Modes          = _spirit.Parameters_EMA_Set_N_Modes
_EMA_Set_N_Modes.argtypes = [ctypes.c_void_p, ctypes.c_int,
                            ctypes.c_int, ctypes.c_int]
_EMA_Set_N_Modes.restype  = None
def set_n_modes(p_state, n_modes, idx_image=-1, idx_chain=-1):
    """Set the number of modes to calculate or use."""
    _EMA_Set_N_Modes(ctypes.c_void_p(p_state), ctypes.c_int(n_modes),
                          ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

_EMA_Set_N_Mode_Follow          = _spirit.Parameters_EMA_Set_N_Mode_Follow
_EMA_Set_N_Mode_Follow.argtypes = [ctypes.c_void_p, ctypes.c_int,
                                    ctypes.c_int, ctypes.c_int]
_EMA_Set_N_Mode_Follow.restype  = None
def set_n_mode_follow(p_state, n_mode, idx_image=-1, idx_chain=-1):
    """Set the index of the mode to use."""
    _EMA_Set_N_Mode_Follow(ctypes.c_void_p(p_state), ctypes.c_int(n_mode),
                          ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

## ---------------------------------- Get ----------------------------------

_EMA_Get_N_Modes          = _spirit.Parameters_EMA_Get_N_Modes
_EMA_Get_N_Modes.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_EMA_Get_N_Modes.restype  = ctypes.c_int
def get_n_modes(p_state, idx_image=-1, idx_chain=-1):
    """Returns the number of modes to calculate or use."""
    return int(_EMA_Get_N_Modes(p_state, ctypes.c_int(idx_image), ctypes.c_int(idx_chain)))

_EMA_Get_N_Mode_Follow          = _spirit.Parameters_EMA_Get_N_Mode_Follow
_EMA_Get_N_Mode_Follow.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_EMA_Get_N_Mode_Follow.restype  = ctypes.c_int
def get_n_mode_follow(p_state, idx_image=-1, idx_chain=-1):
    """Returns the index of the mode to use."""
    return int(_EMA_Get_N_Mode_Follow(p_state, ctypes.c_int(idx_image), ctypes.c_int(idx_chain)))


"""
Monte Carlo (MC)
-------------------------------------------------------------
"""

import spirit.spiritlib as spiritlib
from spirit.io import FILEFORMAT_OVF_TEXT
import ctypes

### Load Library
_spirit = spiritlib.load_spirit_library()

## ---------------------------------- Set ----------------------------------

_MC_Set_Output_Tag          = _spirit.Parameters_MC_Set_Output_Tag
_MC_Set_Output_Tag.argtypes = [ctypes.c_void_p, ctypes.c_char_p,
                                ctypes.c_int, ctypes.c_int]
_MC_Set_Output_Tag.restype  = None
def set_output_tag(p_state, tag, idx_image=-1, idx_chain=-1):
    """Set the tag placed in front of output file names.

    If the tag is "<time>", it will be the date-time of the creation of the state."""
    _MC_Set_Output_Tag(ctypes.c_void_p(p_state), ctypes.c_char_p(tag.encode('utf-8')),
                        ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

_MC_Set_Output_Folder          = _spirit.Parameters_MC_Set_Output_Folder
_MC_Set_Output_Folder.argtypes = [ctypes.c_void_p, ctypes.c_char_p,
                                    ctypes.c_int, ctypes.c_int]
_MC_Set_Output_Folder.restype  = None
def set_output_folder(p_state, folder, idx_image=-1, idx_chain=-1):
    """Set the folder, where output files are placed."""
    _MC_Set_Output_Folder(ctypes.c_void_p(p_state), ctypes.c_char_p(folder.encode('utf-8')),
                        ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

_MC_Set_Output_General          = _spirit.Parameters_MC_Set_Output_General
_MC_Set_Output_General.argtypes = [ctypes.c_void_p, ctypes.c_bool, ctypes.c_bool,
                                    ctypes.c_bool, ctypes.c_int, ctypes.c_int]
_MC_Set_Output_General.restype  = None
def set_output_general(p_state, any=True, initial=False, final=False, idx_image=-1, idx_chain=-1):
    """Set whether to write any output files at all."""
    _MC_Set_Output_General(ctypes.c_void_p(p_state), ctypes.c_bool(any), ctypes.c_bool(initial),
                        ctypes.c_bool(final), ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

_MC_Set_Output_Energy          = _spirit.Parameters_MC_Set_Output_Energy
_MC_Set_Output_Energy.argtypes = [ctypes.c_void_p, ctypes.c_bool, ctypes.c_bool,
                                    ctypes.c_bool, ctypes.c_bool,
                                    ctypes.c_bool, ctypes.c_int, ctypes.c_int]
_MC_Set_Output_Energy.restype  = None
def set_output_energy(p_state, step=False, archive=True, spin_resolved=False, divide_by_nos=True, add_readability_lines=True, idx_image=-1, idx_chain=-1):
    """Set whether to write energy output files.

    - step: whether to write a new file after each set of iterations
    - archive: whether to append to an archive file after each set of iterations
    - spin_resolved: whether to write a file containing the energy of each spin
    - divide_by_nos: whether to divide energies by the number of spins
    - add_readability_lines: whether to separate columns by lines
    """
    _MC_Set_Output_Energy(ctypes.c_void_p(p_state), ctypes.c_bool(step), ctypes.c_bool(archive),
                        ctypes.c_bool(spin_resolved), ctypes.c_bool(divide_by_nos), ctypes.c_bool(add_readability_lines),
                        ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

_MC_Set_Output_Configuration          = _spirit.Parameters_MC_Set_Output_Configuration
_MC_Set_Output_Configuration.argtypes = [ctypes.c_void_p, ctypes.c_bool, ctypes.c_bool,
                                            ctypes.c_int, ctypes.c_int, ctypes.c_int]
_MC_Set_Output_Configuration.restype  = None
def set_output_configuration(p_state, step=False, archive=True, filetype=FILEFORMAT_OVF_TEXT, idx_image=-1, idx_chain=-1):
    """Set whether to write spin configuration output files.

    - step: whether to write a new file after each set of iterations
    - archive: whether to append to an archive file after each set of iterations
    - filetype: the format in which the data is written
    """
    _MC_Set_Output_Configuration(ctypes.c_void_p(p_state), ctypes.c_bool(step), ctypes.c_bool(archive),
                        ctypes.c_int(filetype), ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

_MC_Set_N_Iterations             = _spirit.Parameters_MC_Set_N_Iterations
_MC_Set_N_Iterations.argtypes    = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int,
                                     ctypes.c_int, ctypes.c_int]
_MC_Set_N_Iterations.restype     = None
def set_iterations(p_state, n_iterations, n_iterations_log, idx_image=-1, idx_chain=-1):
    """Set the number of iterations and how often to log and write output.

    - n_iterations: the maximum number of iterations
    - n_iterations_log: the number of iterations after which status is logged and output written
    """
    _MC_Set_N_Iterations(ctypes.c_void_p(p_state), ctypes.c_int(n_iterations),
                          ctypes.c_int(n_iterations_log), ctypes.c_int(idx_image),
                          ctypes.c_int(idx_chain))

_MC_Set_Temperature             = _spirit.Parameters_MC_Set_Temperature
_MC_Set_Temperature.argtypes    = [ctypes.c_void_p, ctypes.c_float, ctypes.c_int, ctypes.c_int]
_MC_Set_Temperature.restype     = None
def set_temperature(p_state, temperature, idx_image=-1, idx_chain=-1):
    """Set the global base temperature [K]."""
    _MC_Set_Temperature(ctypes.c_void_p(p_state), ctypes.c_float(temperature),
                         ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

_MC_Set_Metropolis_Cone             = _spirit.Parameters_MC_Set_Metropolis_Cone
_MC_Set_Metropolis_Cone.argtypes    = [ctypes.c_void_p, ctypes.c_bool, ctypes.c_float,
                                            ctypes.c_bool, ctypes.c_float, ctypes.c_int, ctypes.c_int]
_MC_Set_Metropolis_Cone.restype     = None
def set_metropolis_cone(p_state, use_cone=True, cone_angle=40, use_adaptive_cone=True, target_acceptance_ratio=0.5, idx_image=-1, idx_chain=-1):
    """Configure the Metropolis parameters.

    - use_cone: whether to displace the spins within a cone (otherwise: on the entire unit sphere)
    - cone_angle: the opening angle within which the spin is placed
    - use_adaptive_cone: automatically adapt the cone angle to achieve the set acceptance ratio
    - target_acceptance_ratio: target acceptance ratio for the adaptive cone algorithm
    """
    _MC_Set_Metropolis_Cone(p_state, ctypes.c_bool(use_cone), ctypes.c_float(cone_angle), ctypes.c_bool(use_adaptive_cone), ctypes.c_float(target_acceptance_ratio), idx_image, idx_chain)

## ---------------------------------- Get ----------------------------------

_MC_Get_N_Iterations             = _spirit.Parameters_MC_Get_N_Iterations
_MC_Get_N_Iterations.argtypes    = [ctypes.c_void_p, ctypes.POINTER( ctypes.c_int ),
                                     ctypes.POINTER( ctypes.c_int ), ctypes.c_int, ctypes.c_int]
_MC_Get_N_Iterations.restype     = None
def get_iterations(p_state, idx_image=-1, idx_chain=-1):
    """Returns the maximum number of iterations and the step size."""
    n_iterations = ctypes.c_int()
    n_iterations_log = ctypes.c_int()
    _MC_Get_N_Iterations(p_state, ctypes.pointer(n_iterations), ctypes.pointer(n_iterations_log),
                          ctypes.c_int(idx_image), ctypes.c_int(idx_chain))
    return int(n_iterations.value), int(n_iterations_log.value)

_MC_Get_Temperature             = _spirit.Parameters_MC_Get_Temperature
_MC_Get_Temperature.argtypes    = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_MC_Get_Temperature.restype     = ctypes.c_float
def get_temperature(p_state, idx_image=-1, idx_chain=-1):
    """Returns the global base temperature [K]."""
    return float(_MC_Get_Temperature(ctypes.c_void_p(p_state), ctypes.c_int(idx_image),
                                             ctypes.c_int(idx_chain)))

_MC_Get_Metropolis_Cone             = _spirit.Parameters_MC_Get_Metropolis_Cone
_MC_Get_Metropolis_Cone.argtypes    = [ ctypes.c_void_p,
                                        ctypes.POINTER(ctypes.c_bool), ctypes.POINTER(ctypes.c_float),
                                        ctypes.POINTER(ctypes.c_bool), ctypes.POINTER(ctypes.c_float),
                                        ctypes.c_int, ctypes.c_int ]
_MC_Get_Metropolis_Cone.restype     = None
def get_metropolis_cone(p_state, idx_image=-1, idx_chain=-1):
    """Returns the Metropolis algorithm configuration.

    - whether the spins are displaced within a cone (otherwise: on the entire unit sphere)
    - the opening angle within which the spin is placed
    - whether the cone angle is automatically adapted to achieve the set acceptance ratio
    - target acceptance ratio for the adaptive cone algorithm
    """
    use_cone = ctypes.c_bool()
    cone_angle = ctypes.c_float()
    use_adaptive_cone = ctypes.c_bool()
    target_acceptance_ratio = ctypes.c_float()
    _MC_Get_Metropolis_Cone(ctypes.c_void_p(p_state),
                ctypes.pointer(use_cone), ctypes.pointer(cone_angle),
                ctypes.pointer(use_adaptive_cone), ctypes.pointer(target_acceptance_ratio),
                ctypes.c_int(idx_image), ctypes.c_int(idx_chain))
    return bool(use_cone), float(cone_angle), bool(use_adaptive_cone), float(target_acceptance_ratio)


"""
Quantities
====================
"""

import spirit.spiritlib as spiritlib
import spirit.system as system
from spirit.scalar import scalar
import ctypes

import numpy as np

### Load Library
_spirit = spiritlib.load_spirit_library()

_Get_Magnetization          = _spirit.Quantity_Get_Magnetization
_Get_Magnetization.argtypes = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_float),
                               ctypes.c_int, ctypes.c_int]
_Get_Magnetization.restype  = None
def get_magnetization(p_state, idx_image=-1, idx_chain=-1):
    """Calculates and returns the average magnetization of the system as
    an array of shape (3).
    """
    magnetization = (3*ctypes.c_float)()
    _Get_Magnetization(ctypes.c_void_p(p_state), magnetization,
                       ctypes.c_int(idx_image), ctypes.c_int(idx_chain))
    return [float(i) for i in magnetization]


### Temporary info function for MMF
_Get_MinimumMode            = _spirit.Quantity_Get_Grad_Force_MinimumMode
_Get_MinimumMode.argtypes   = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float),
                                ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.c_int, ctypes.c_int]
_Get_MinimumMode.restype    = None
def get_mmf_info(p_state, idx_image=-1, idx_chain=-1):
    """Returns a set of MMF information, meant mostly for testing or debugging.

    - numpy.array_view of shape (NOS, 3) of the energy gradient
    - the lowest eigenvalue
    - numpy.array_view of shape (NOS, 3) of the eigenmode
    - numpy.array_view of shape (NOS, 3) of the force
    """
    nos = system.get_nos(p_state, idx_image, idx_chain)

    ArrayType = ctypes.c_float*(3*nos)

    MM = []*(3*nos)
    _MM = ArrayType(*MM)
    FF = []*(3*nos)
    _FF = ArrayType(*FF)
    GG = []*(3*nos)
    _GG = ArrayType(*FF)
    ev = []*1
    _eval = ArrayType(*ev)

    _Get_MinimumMode(p_state, _GG, _eval, _MM, _FF, idx_image, idx_chain)

    # array_pointer = ctypes.cast(_MM, ctypes.POINTER(ArrayType))
    # array = np.frombuffer(array_pointer.contents)

    MMM = np.array(_MM)
    FFF = np.array(_FF)
    GGG = np.array(_GG)
    array_view_mode = MMM.view()
    array_view_mode.shape = (nos, 3)
    array_view_force = FFF.view()
    array_view_force.shape = (nos, 3)
    array_view_grad = GGG.view()
    array_view_grad.shape = (nos, 3)

    return array_view_grad, _eval[0], array_view_mode, array_view_force


_Get_Topological_Charge          = _spirit.Quantity_Get_Topological_Charge
_Get_Topological_Charge.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Get_Topological_Charge.restype  = ctypes.c_float
def get_topological_charge(p_state, idx_image=-1, idx_chain=-1):
    """Calculates and returns the total topological charge of 2D systems.

    Note that the charge can take unphysical non-integer values for open boundaries,
    because it is not well-defined in this case.

    Returns 0 for systems of other dimensionality.
    """
    return float(_Get_Topological_Charge(ctypes.c_void_p(p_state),
                       ctypes.c_int(idx_image), ctypes.c_int(idx_chain)))


"""
Geometry
====================

Change or get info on the current geometrical configuration, e.g.
number of cells in the three crystal translation directions.
"""

import spirit.spiritlib as spiritlib
import ctypes

### Load Library
_spirit = spiritlib.load_spirit_library()

### Imports
from spirit.scalar import scalar
from spirit import system

import numpy as np

### Bravais lattice types
BRAVAIS_LATTICE_IRREGULAR   = 0
BRAVAIS_LATTICE_RECTILINEAR = 1
BRAVAIS_LATTICE_SC          = 2
BRAVAIS_LATTICE_HEX2D       = 3
BRAVAIS_LATTICE_HEX2D_60    = 4
BRAVAIS_LATTICE_HEX2D_120   = 5
BRAVAIS_LATTICE_HCP         = 6
BRAVAIS_LATTICE_BCC         = 7
BRAVAIS_LATTICE_FCC         = 8

### ---------------------------------- Set ----------------------------------

_Set_Bravais_Lattice_Type          = _spirit.Geometry_Set_Bravais_Lattice_Type
_Set_Bravais_Lattice_Type.argtypes = [ctypes.c_void_p, ctypes.c_int]
_Set_Bravais_Lattice_Type.restype  = None
def set_bravais_lattice_type(p_state, lattice_type, idx_image=-1, idx_chain=-1):
    """Set the bravais vectors to a pre-defined lattice type:

    - sc:       simple cubic
    - bcc:      body centered cubic
    - fcc:      face centered cubic
    - hex2d:    hexagonal (120deg)
    - hed2d120: hexagonal (120deg)
    - hex2d60:  hexagonal (60deg)
    """
    _Set_Bravais_Lattice_Type(ctypes.c_void_p(p_state), ctypes.c_int(lattice_type))

_Set_N_Cells          = _spirit.Geometry_Set_N_Cells
_Set_N_Cells.argtypes = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_int)]
_Set_N_Cells.restype  = None
def set_n_cells(p_state, n_cells=[1, 1, 1], idx_image=-1, idx_chain=-1):
    """Set the number of basis cells along the three bravais vectors."""
    vec3 = ctypes.c_int * 3
    _Set_N_Cells(ctypes.c_void_p(p_state), vec3(*n_cells))

_Set_mu_s             = _spirit.Geometry_Set_mu_s
_Set_mu_s.argtypes    = [ctypes.c_void_p, ctypes.c_float,
                          ctypes.c_int, ctypes.c_int]
_Set_mu_s.restype     = None
def set_mu_s(p_state, mu_s, idx_image=-1, idx_chain=-1):
    """Set the magnetic moment of all atoms."""
    _Set_mu_s(ctypes.c_void_p(p_state), ctypes.c_float(mu_s),
              ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

_Set_Cell_Atom_Types          = _spirit.Geometry_Set_Cell_Atom_Types
_Set_Cell_Atom_Types.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.POINTER(ctypes.c_float)]
_Set_Cell_Atom_Types.restype  = None
def set_cell_atom_types(p_state, atom_types, idx_image=-1, idx_chain=-1):
    """Set the atom types of the atoms in the basis cell."""
    n = len(atom_types)
    vec = ctypes.c_int * n
    _Set_Cell_Atom_Types(ctypes.c_void_p(p_state), ctypes.c_int(n), vec(*atom_types))

_Set_Bravais_Vectors             = _spirit.Geometry_Set_Bravais_Vectors
_Set_Bravais_Vectors.argtypes    = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_float),
                                    ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float)]
_Set_Bravais_Vectors.restype     = None
def set_bravais_vectors(p_state, ta=[1.0, 0.0, 0.0], tb=[0.0, 1.0, 0.0], tc=[0.0, 0.0, 1.0], idx_image=-1, idx_chain=-1):
    """Manually specify the bravais vectors."""
    vec3 = ctypes.c_float * 3
    _Set_Bravais_Vectors(ctypes.c_void_p(p_state), vec3(ta), vec3(tb), vec3(tc))

_Set_Lattice_Constant             = _spirit.Geometry_Set_Lattice_Constant
_Set_Lattice_Constant.argtypes    = [ctypes.c_void_p, ctypes.c_float]
_Set_Lattice_Constant.restype     = None
def set_lattice_constant(p_state, lattice_constant, idx_image=-1, idx_chain=-1):
    """Set the global lattice scaling constant."""
    _Set_Lattice_Constant(p_state, ctypes.c_float(lattice_constant))

### ---------------------------------- Get ----------------------------------

_Get_Bounds          = _spirit.Geometry_Get_Bounds
_Get_Bounds.argtypes = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_float),
                        ctypes.POINTER(ctypes.c_float), ctypes.c_int, ctypes.c_int]
_Get_Bounds.restype  = None
def get_bounds(p_state, idx_image=-1, idx_chain=-1):
    """Get the bounds of the system in global coordinates.

    Returns two arrays of shape (3) containing minimum and maximum bounds respectively.
    """
    _min = (3*ctypes.c_float)()
    _max = (3*ctypes.c_float)()
    _Get_Bounds(ctypes.c_void_p(p_state), _min, _max,
                ctypes.c_int(idx_image), ctypes.c_int(idx_chain))
    return [_min[i] for i in range(3)], [_max[i] for i in range(3)]

_Get_Center          = _spirit.Geometry_Get_Center
_Get_Center.argtypes = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_float), ctypes.c_int, ctypes.c_int]
_Get_Center.restype  = None
def get_center(p_state, idx_image=-1, idx_chain=-1):
    """Get the center of the system in global coordinates.

    Returns an array of shape (3).
    """
    _center = (3*ctypes.c_float)()
    _Get_Center(ctypes.c_void_p(p_state), _center, ctypes.c_int(idx_image), ctypes.c_int(idx_chain))
    return [_center[i] for i in range(3)]

_Get_Bravais_Lattice_Type          = _spirit.Geometry_Get_Bravais_Lattice_Type
_Get_Bravais_Lattice_Type.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Get_Bravais_Lattice_Type.restype  = ctypes.c_int
def get_bravais_lattice_type(p_state, idx_image=-1, idx_chain=-1):
    """Get the bravais lattice type corresponding to one of the integers defined above."""
    return int(_Get_Bravais_Lattice_Type(ctypes.c_void_p(p_state), ctypes.c_int(idx_image),
                                 ctypes.c_int(idx_chain)))

_Get_Bravais_Vectors          = _spirit.Geometry_Get_Bravais_Vectors
_Get_Bravais_Vectors.argtypes = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_float),
                               ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float),
                               ctypes.c_int, ctypes.c_int]
_Get_Bravais_Vectors.restype  = None
def get_bravais_vectors(p_state, idx_image=-1, idx_chain=-1):
    """Get the Bravais vectors.

    Returns three arrays of shape (3).
    """
    _a = (3*ctypes.c_float)()
    _b = (3*ctypes.c_float)()
    _c = (3*ctypes.c_float)()
    _Get_Bravais_Vectors(ctypes.c_void_p(p_state), _a, _b, _c,
                       ctypes.c_int(idx_image), ctypes.c_int(idx_chain))
    return [a for a in _a], [b for b in _b], [c for c in _c]

_Get_N_Cells          = _spirit.Geometry_Get_N_Cells
_Get_N_Cells.argtypes = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_int), ctypes.c_int, ctypes.c_int]
_Get_N_Cells.restype  = None
def get_n_cells(p_state, idx_image=-1, idx_chain=-1):
    """Get the number of basis cells along the three bravais vectors.

    Returns an array of shape (3).
    """
    n_cells = (3*ctypes.c_int)()
    _Get_N_Cells(ctypes.c_void_p(p_state), n_cells, ctypes.c_int(idx_image), ctypes.c_int(idx_chain))
    return [n for n in n_cells]

_Get_Dimensionality          = _spirit.Geometry_Get_Dimensionality
_Get_Dimensionality.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Get_Dimensionality.restype  = ctypes.c_int
def get_dimensionality(p_state, idx_image=-1, idx_chain=-1):
    """Get the dimensionality of the geometry."""
    return int(_Get_Dimensionality(ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)))

### Get Pointer to Spin Positions
# NOTE: Changing the values of the array_view one can alter the value of the data of the state
_Get_Positions            = _spirit.Geometry_Get_Positions
_Get_Positions.argtypes   = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Get_Positions.restype    = ctypes.POINTER(scalar)
def get_positions(p_state, idx_image=-1, idx_chain=-1):
    """Returns a `numpy.array_view` of shape (NOS, 3) with the components of each spins position.

    Changing the contents of this array_view will have direct effect on the state and should not be done.
    """
    nos = system.get_nos(p_state, idx_image, idx_chain)
    ArrayType = scalar*3*nos
    Data = _Get_Positions(ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain))
    array_pointer = ctypes.cast(Data, ctypes.POINTER(ArrayType))
    array = np.frombuffer(array_pointer.contents, dtype=scalar)
    array_view = array.view()
    array_view.shape = (nos, 3)
    return array_view

_Get_N_Cell_Atoms          = _spirit.Geometry_Get_N_Cell_Atoms
_Get_N_Cell_Atoms.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Get_N_Cell_Atoms.restype  = ctypes.c_int
def get_n_cell_atoms(p_state, idx_image=-1, idx_chain=-1):
    """Get the number of atoms in the basis cell."""
    return int(_Get_N_Cell_Atoms(ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)))

### Get Pointer to atom types
# NOTE: Changing the values of the array_view one can alter the value of the data of the state
_Get_Atom_Types            = _spirit.Geometry_Get_Atom_Types
_Get_Atom_Types.argtypes   = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Get_Atom_Types.restype    = ctypes.POINTER(ctypes.c_int)
def get_atom_types(p_state, idx_image=-1, idx_chain=-1):
    """Get the types of all atoms as a `numpy.array_view` of shape (NOS).

    If e.g. disorder is activated, this allows to view and manipulate the types of individual atoms.
    """
    nos = system.get_nos(p_state, idx_image, idx_chain)
    ArrayType = ctypes.c_int*nos
    Data = _Get_Atom_Types(ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain))
    array_pointer = ctypes.cast(Data, ctypes.POINTER(ArrayType))
    array = np.frombuffer(array_pointer.contents, dtype=ctypes.c_int)
    array_view = array.view()
    return array_view


"""
Landau-Lifshitz-Gilbert (LLG)
-------------------------------------------------------------
"""

import spirit.spiritlib as spiritlib
from spirit.io import FILEFORMAT_OVF_TEXT
import ctypes

### Load Library
_spirit = spiritlib.load_spirit_library()

### ---------------------------------- Set ----------------------------------

_LLG_Set_Output_Tag          = _spirit.Parameters_LLG_Set_Output_Tag
_LLG_Set_Output_Tag.argtypes = [ctypes.c_void_p, ctypes.c_char_p,
                                ctypes.c_int, ctypes.c_int]
_LLG_Set_Output_Tag.restype  = None
def set_output_tag(p_state, tag, idx_image=-1, idx_chain=-1):
    """Set the tag placed in front of output file names.

    If the tag is "<time>", it will be the date-time of the creation of the state."""
    _LLG_Set_Output_Tag(ctypes.c_void_p(p_state), ctypes.c_char_p(tag.encode('utf-8')),
                        ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

_LLG_Set_Output_Folder          = _spirit.Parameters_LLG_Set_Output_Folder
_LLG_Set_Output_Folder.argtypes = [ctypes.c_void_p, ctypes.c_char_p,
                                    ctypes.c_int, ctypes.c_int]
_LLG_Set_Output_Folder.restype  = None
def set_output_folder(p_state, folder, idx_image=-1, idx_chain=-1):
    """Set the folder, where output files are placed."""
    _LLG_Set_Output_Folder(ctypes.c_void_p(p_state), ctypes.c_char_p(folder.encode('utf-8')),
                        ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

_LLG_Set_Output_General          = _spirit.Parameters_LLG_Set_Output_General
_LLG_Set_Output_General.argtypes = [ctypes.c_void_p, ctypes.c_bool, ctypes.c_bool,
                                    ctypes.c_bool, ctypes.c_int, ctypes.c_int]
_LLG_Set_Output_General.restype  = None
def set_output_general(p_state, any=True, initial=False, final=False, idx_image=-1, idx_chain=-1):
    """Set whether to write any output files at all."""
    _LLG_Set_Output_General(ctypes.c_void_p(p_state), ctypes.c_bool(any), ctypes.c_bool(initial),
                        ctypes.c_bool(final), ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

_LLG_Set_Output_Energy          = _spirit.Parameters_LLG_Set_Output_Energy
_LLG_Set_Output_Energy.argtypes = [ctypes.c_void_p, ctypes.c_bool, ctypes.c_bool,
                                    ctypes.c_bool, ctypes.c_bool,
                                    ctypes.c_bool, ctypes.c_int, ctypes.c_int]
_LLG_Set_Output_Energy.restype  = None
def set_output_energy(p_state, step=False, archive=True, spin_resolved=False, divide_by_nos=True, add_readability_lines=True, idx_image=-1, idx_chain=-1):
    """Set whether to write energy output files.

    - step: whether to write a new file after each set of iterations
    - archive: whether to append to an archive file after each set of iterations
    - spin_resolved: whether to write a file containing the energy of each spin
    - divide_by_nos: whether to divide energies by the number of spins
    - add_readability_lines: whether to separate columns by lines
    """
    _LLG_Set_Output_Energy(ctypes.c_void_p(p_state), ctypes.c_bool(step), ctypes.c_bool(archive),
                        ctypes.c_bool(spin_resolved), ctypes.c_bool(divide_by_nos), ctypes.c_bool(add_readability_lines),
                        ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

_LLG_Set_Output_Configuration          = _spirit.Parameters_LLG_Set_Output_Configuration
_LLG_Set_Output_Configuration.argtypes = [ctypes.c_void_p, ctypes.c_bool, ctypes.c_bool,
                                            ctypes.c_int, ctypes.c_int, ctypes.c_int]
_LLG_Set_Output_Configuration.restype  = None
def set_output_configuration(p_state, step=False, archive=True, filetype=FILEFORMAT_OVF_TEXT, idx_image=-1, idx_chain=-1):
    """Set whether to write spin configuration output files.

    - step: whether to write a new file after each set of iterations
    - archive: whether to append to an archive file after each set of iterations
    - filetype: the format in which the data is written
    """
    _LLG_Set_Output_Configuration(ctypes.c_void_p(p_state), ctypes.c_bool(step), ctypes.c_bool(archive),
                        ctypes.c_int(filetype), ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

_LLG_Set_N_Iterations             = _spirit.Parameters_LLG_Set_N_Iterations
_LLG_Set_N_Iterations.argtypes    = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int,
                                     ctypes.c_int, ctypes.c_int]
_LLG_Set_N_Iterations.restype     = None
def set_iterations(p_state, n_iterations, n_iterations_log, idx_image=-1, idx_chain=-1):
    """Set the number of iterations and how often to log and write output.

    - n_iterations: the maximum number of iterations
    - n_iterations_log: the number of iterations after which status is logged and output written
    """
    _LLG_Set_N_Iterations(ctypes.c_void_p(p_state), ctypes.c_int(n_iterations),
                          ctypes.c_int(n_iterations_log), ctypes.c_int(idx_image),
                          ctypes.c_int(idx_chain))

_LLG_Set_Direct_Minimization            = _spirit.Parameters_LLG_Set_Direct_Minimization
_LLG_Set_Direct_Minimization.argtypes   = [ctypes.c_void_p, ctypes.c_bool,
                                           ctypes.c_int, ctypes.c_int ]
_LLG_Set_Direct_Minimization.restype    = None
def set_direct_minimization(p_state, use_minimization, idx_image=-1, idx_chain=-1):
    """Set whether to minimise the energy without precession.

    This only influences dynamics solvers, which will then perform pseudodynamics,
    simulating only the damping part of the LLG equation.
    """
    _LLG_Set_Direct_Minimization(ctypes.c_void_p(p_state), ctypes.c_bool(use_minimization),
                                 ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

_LLG_Set_Convergence            = _spirit.Parameters_LLG_Set_Convergence
_LLG_Set_Convergence.argtypes   = [ctypes.c_void_p, ctypes.c_float, ctypes.c_int, ctypes.c_int ]
_LLG_Set_Convergence.restype    = None
def set_convergence(p_state, convergence, idx_image=-1, idx_chain=-1):
    """Set the convergence limit.

    When the maximum absolute component value of the force drops below this value,
    the calculation is considered converged and will stop.
    """
    _LLG_Set_Convergence(ctypes.c_void_p(p_state), ctypes.c_float(convergence),
                         ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

_LLG_Set_Time_Step             = _spirit.Parameters_LLG_Set_Time_Step
_LLG_Set_Time_Step.argtypes    = [ctypes.c_void_p, ctypes.c_float, ctypes.c_int, ctypes.c_int]
_LLG_Set_Time_Step.restype     = None
def set_timestep(p_state, dt, idx_image=-1, idx_chain=-1):
    """Set the time step [ps] for the calculation."""
    _LLG_Set_Time_Step(p_state, ctypes.c_float(dt), idx_image, idx_chain)

_LLG_Set_Damping             = _spirit.Parameters_LLG_Set_Damping
_LLG_Set_Damping.argtypes    = [ctypes.c_void_p, ctypes.c_float, ctypes.c_int, ctypes.c_int]
_LLG_Set_Damping.restype     = None
def set_damping(p_state, damping, idx_image=-1, idx_chain=-1):
    """Set the Gilbert damping parameter [unitless]."""
    _LLG_Set_Damping(ctypes.c_void_p(p_state), ctypes.c_float(damping),
                     ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

_LLG_Set_STT             = _spirit.Parameters_LLG_Set_STT
_LLG_Set_STT.argtypes    = [ctypes.c_void_p, ctypes.c_bool, ctypes.c_float,
                            ctypes.POINTER(3*ctypes.c_float), ctypes.c_int, ctypes.c_int]
_LLG_Set_STT.restype     = None
def set_stt(p_state, use_gradient, magnitude, direction, idx_image=-1, idx_chain=-1):
    """Set the spin current configuration.

    - use_gradient: `True`: use the spatial gradient, `False`: monolayer approximation
    - magnitude: current strength
    - direction: current direction or polarisation direction, array of shape (3)
    """
    vec3 = ctypes.c_float * 3
    direction = vec3(*direction)
    ctypes.cast( direction, ctypes.POINTER(vec3))
    _LLG_Set_STT(ctypes.c_void_p(p_state), ctypes.c_bool(use_gradient),
                 ctypes.c_float(magnitude), direction, ctypes.c_int(idx_image),
                 ctypes.c_int(idx_chain))

_LLG_Set_Temperature             = _spirit.Parameters_LLG_Set_Temperature
_LLG_Set_Temperature.argtypes    = [ctypes.c_void_p, ctypes.c_float, ctypes.c_int, ctypes.c_int]
_LLG_Set_Temperature.restype     = None
_LLG_Set_Temperature_Gradient             = _spirit.Parameters_LLG_Set_Temperature_Gradient
_LLG_Set_Temperature_Gradient.argtypes    = [ctypes.c_void_p, ctypes.c_float, ctypes.POINTER(3*ctypes.c_float),
                                             ctypes.c_int, ctypes.c_int]
_LLG_Set_Temperature_Gradient.restype     = None
def set_temperature(p_state, temperature, gradient_inclination=0, gradient_direction=[1.0,0.0,0.0], idx_image=-1, idx_chain=-1):
    """Set the temperature configuration.

    - temperature: homogeneous base temperature [K]
    - gradient_inclination: inclination of the temperature gradient [K/a]
    - gradient_direction: direction of the temperature gradient, array of shape (3)
    """
    _LLG_Set_Temperature(ctypes.c_void_p(p_state), ctypes.c_float(temperature),
                         ctypes.c_int(idx_image), ctypes.c_int(idx_chain))
    vec3 = ctypes.c_float * 3
    gradient_direction = vec3(*gradient_direction)
    ctypes.cast( gradient_direction, ctypes.POINTER(vec3))
    _LLG_Set_Temperature_Gradient(ctypes.c_void_p(p_state), ctypes.c_float(gradient_inclination), gradient_direction,
                                  ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

### ---------------------------------- Get ----------------------------------

_LLG_Get_N_Iterations             = _spirit.Parameters_LLG_Get_N_Iterations
_LLG_Get_N_Iterations.argtypes    = [ctypes.c_void_p, ctypes.POINTER( ctypes.c_int ),
                                     ctypes.POINTER( ctypes.c_int ), ctypes.c_int, ctypes.c_int]
_LLG_Get_N_Iterations.restype     = None
def get_iterations(p_state, idx_image=-1, idx_chain=-1):
    """Returns the maximum number of iterations and the step size."""
    n_iterations = ctypes.c_int()
    n_iterations_log = ctypes.c_int()
    _LLG_Get_N_Iterations(p_state, ctypes.pointer(n_iterations), ctypes.pointer(n_iterations_log),
                          ctypes.c_int(idx_image), ctypes.c_int(idx_chain))
    return int(n_iterations.value), int(n_iterations_log.value)

_LLG_Get_Direct_Minimization            = _spirit.Parameters_LLG_Get_Direct_Minimization
_LLG_Get_Direct_Minimization.argtypes   = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int ]
_LLG_Get_Direct_Minimization.restype    = ctypes.c_bool
def get_direct_minimization(p_state, idx_image=-1, idx_chain=-1):
    """Returns whether only energy minimisation will be performed."""
    return float(_LLG_Get_Direct_Minimization(ctypes.c_void_p(p_state),
                                              ctypes.c_int(idx_image), ctypes.c_int(idx_chain)))

_LLG_Get_Convergence            = _spirit.Parameters_LLG_Get_Convergence
_LLG_Get_Convergence.argtypes   = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int ]
_LLG_Get_Convergence.restype    = ctypes.c_float
def get_convergence(p_state, idx_image=-1, idx_chain=-1):
    """Returns the convergence value."""
    return float(_LLG_Get_Convergence(ctypes.c_void_p(p_state), ctypes.c_int(idx_image),
                                      ctypes.c_int(idx_chain)))

_LLG_Get_Time_Step             = _spirit.Parameters_LLG_Get_Time_Step
_LLG_Get_Time_Step.argtypes    = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_LLG_Get_Time_Step.restype     = ctypes.c_float
def get_timestep(p_state, idx_image=-1, idx_chain=-1):
    """Returns the time step [ps]."""
    return float(_LLG_Get_Time_Step(ctypes.c_void_p(p_state), ctypes.c_int(idx_image),
                                    ctypes.c_int(idx_chain)))

_LLG_Get_Damping             = _spirit.Parameters_LLG_Get_Damping
_LLG_Get_Damping.argtypes    = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_LLG_Get_Damping.restype     = ctypes.c_float
def get_damping(p_state, idx_image=-1, idx_chain=-1):
    """Returns the Gilbert damping parameter."""
    return float(_LLG_Get_Damping(ctypes.c_void_p(p_state), ctypes.c_int(idx_image),
                            ctypes.c_int(idx_chain)))

_LLG_Get_STT             = _spirit.Parameters_LLG_Get_STT
_LLG_Get_STT.argtypes    = [ctypes.c_void_p, ctypes.POINTER( ctypes.c_bool ),
                            ctypes.POINTER( ctypes.c_float ), ctypes.POINTER(ctypes.c_float),
                            ctypes.c_int, ctypes.c_int]
_LLG_Get_STT.restype     = None
def get_stt(p_state, idx_image=-1, idx_chain=-1):
    """Returns the spin current configuration.

    - magnitude
    - direction, array of shape (3)
    - whether the spatial gradient is used
    """
    direction = (3*ctypes.c_float)()
    use_gradient = ctypes.c_bool()
    magnitude = ctypes.c_float()
    _LLG_Get_STT(ctypes.c_void_p(p_state), ctypes.pointer(use_gradient),
                 ctypes.pointer(magnitude), direction, ctypes.c_int(idx_image),
                 ctypes.c_int(idx_chain))
    return float(magnitude.value), [float(i) for i in direction], bool(use_gradient)

_LLG_Get_Temperature             = _spirit.Parameters_LLG_Get_Temperature
_LLG_Get_Temperature.argtypes    = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_LLG_Get_Temperature.restype     = ctypes.c_float
_LLG_Get_Temperature_Gradient             = _spirit.Parameters_LLG_Get_Temperature_Gradient
_LLG_Get_Temperature_Gradient.argtypes    = [ctypes.c_void_p, ctypes.POINTER( ctypes.c_float ),
                                             ctypes.POINTER(ctypes.c_float), ctypes.c_int, ctypes.c_int]
_LLG_Get_Temperature_Gradient.restype     = None
def get_temperature(p_state, idx_image=-1, idx_chain=-1):
    """Returns the temperature configuration.

    - global base temperature [K]
    - inclination of the temperature gradient [K/a]
    - direction of the temperature gradient, array of shape (3)
    """
    temperature = float(_LLG_Get_Temperature(ctypes.c_void_p(p_state), ctypes.c_int(idx_image),
                                             ctypes.c_int(idx_chain)))
    gradient_inclination = ctypes.c_float()
    gradient_direction = (3*ctypes.c_float)()
    _LLG_Get_Temperature_Gradient(ctypes.c_void_p(p_state), ctypes.pointer(gradient_inclination),
                                  gradient_direction, ctypes.c_int(idx_image), ctypes.c_int(idx_chain))
    return temperature, gradient_inclination, gradient_direction



"""
Configuration
====================

Set various spin configurations, such as homogeneous domains, spirals or skyrmions.
"""

import spirit.spiritlib as spiritlib
import ctypes

### Load Library
_spirit = spiritlib.load_spirit_library()

_Domain             = _spirit.Configuration_Domain
_Domain.argtypes    = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_float),
                       ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float),
                       ctypes.c_float, ctypes.c_float, ctypes.c_bool, ctypes.c_int, ctypes.c_int]
_Domain.restype     = None
def domain(p_state, dir, pos=[0,0,0], border_rectangular=[-1,-1,-1], border_cylindrical=-1,
           border_spherical=-1, inverted=False, idx_image=-1, idx_chain=-1):
    """Set a domain (homogeneous) configuration."""
    vec3 = ctypes.c_float * 3
    _Domain(ctypes.c_void_p(p_state), vec3(*dir), vec3(*pos), vec3(*border_rectangular),
            ctypes.c_float(border_cylindrical), ctypes.c_float(border_spherical),
            ctypes.c_bool(inverted), ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

_PlusZ             = _spirit.Configuration_PlusZ
_PlusZ.argtypes    = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_float),
                      ctypes.POINTER(ctypes.c_float), ctypes.c_float, ctypes.c_float, ctypes.c_bool,
                      ctypes.c_int, ctypes.c_int]
_PlusZ.restype     = None
def plus_z(p_state, pos=[0.0,0.0,0.0], border_rectangular=[-1.0,-1.0,-1.0], border_cylindrical=-1.0,
          border_spherical=-1.0, inverted=False, idx_image=-1, idx_chain=-1):
    """Set a +z (homogeneous) configuration."""
    vec3 = ctypes.c_float * 3
    _PlusZ(ctypes.c_void_p(p_state), vec3(*pos), vec3(*border_rectangular),
           ctypes.c_float(border_cylindrical), ctypes.c_float(border_spherical),
           ctypes.c_bool(inverted), ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

_MinusZ             = _spirit.Configuration_MinusZ
_MinusZ.argtypes    = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_float),
                       ctypes.POINTER(ctypes.c_float), ctypes.c_float, ctypes.c_float,
                       ctypes.c_bool, ctypes.c_int, ctypes.c_int]
_MinusZ.restype     = None
def minus_z(p_state, pos=[0,0,0], border_rectangular=[-1,-1,-1], border_cylindrical=-1,
          border_spherical=-1, inverted=False, idx_image=-1, idx_chain=-1):
    """Set a -z (homogeneous) configuration."""
    vec3 = ctypes.c_float * 3
    _MinusZ(ctypes.c_void_p(p_state), vec3(*pos), vec3(*border_rectangular),
            ctypes.c_float(border_cylindrical), ctypes.c_float(border_spherical),
            ctypes.c_bool(inverted), ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

_Random             = _spirit.Configuration_Random
_Random.argtypes    = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_float),
                       ctypes.POINTER(ctypes.c_float), ctypes.c_float, ctypes.c_float,
                       ctypes.c_bool, ctypes.c_bool, ctypes.c_int, ctypes.c_int]
_Random.restype     = None
def random(p_state, pos=[0,0,0], border_rectangular=[-1,-1,-1], border_cylindrical=-1,
           border_spherical=-1, inverted=False, idx_image=-1, idx_chain=-1):
    """Distribute all spins randomly on the unit sphere."""
    vec3 = ctypes.c_float * 3
    _Random(ctypes.c_void_p(p_state), vec3(*pos), vec3(*border_rectangular),
            ctypes.c_float(border_cylindrical), ctypes.c_float(border_spherical),
            ctypes.c_bool(inverted), False, ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

_Add_Noise_Temperature             = _spirit.Configuration_Add_Noise_Temperature
_Add_Noise_Temperature.argtypes    = [ctypes.c_void_p, ctypes.c_float,
                                      ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float),
                                      ctypes.c_float, ctypes.c_float, ctypes.c_bool,
                                      ctypes.c_int, ctypes.c_int]
_Add_Noise_Temperature.restype     = None
def add_noise(p_state, temperature, pos=[0,0,0], border_rectangular=[-1,-1,-1],
                          border_cylindrical=-1, border_spherical=-1, inverted=False,
                          idx_image=-1, idx_chain=-1):
    """Add temperature-scaled random noise to configuration."""
    vec3 = ctypes.c_float * 3
    _Add_Noise_Temperature(ctypes.c_void_p(p_state), ctypes.c_float(temperature), vec3(*pos),
                           vec3(*border_rectangular), ctypes.c_float(border_cylindrical),
                           ctypes.c_float(border_spherical), ctypes.c_bool(inverted),
                           ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

_Skyrmion             = _spirit.Configuration_Skyrmion
_Skyrmion.argtypes    = [ctypes.c_void_p, ctypes.c_float, ctypes.c_float, ctypes.c_float,
                         ctypes.c_bool, ctypes.c_bool, ctypes.c_bool, ctypes.POINTER(ctypes.c_float),
                         ctypes.POINTER(ctypes.c_float), ctypes.c_float, ctypes.c_float, ctypes.c_bool,
                         ctypes.c_int, ctypes.c_int]
_Skyrmion.restype     = None
def skyrmion(p_state, radius, order=1, phase=1, up_down=False, achiral=False, right_left=False,
             pos=[0,0,0], border_rectangular=[-1,-1,-1], border_cylindrical=-1, border_spherical=-1,
             inverted=False, idx_image=-1, idx_chain=-1):
    """Set a skyrmion configuration.
    
    - radius: the extent of the skyrmion, at which it points approximately upwards
    - order: the number of twists along a circle cutting the skyrmion
    - phase: 0 corresponds to a Neel skyrmion, -90 to a Bloch skyrmion
    - up_down: if `True`, the z-orientation is inverted
    - achiral: if `True`, the topological charge is inverted
    - right_left: if `True`, the in-plane rotation is inverted

    The skyrmion only extends up to `radius`, meaning that `border_cylindrical` is
    not usually necessary.
    """
    vec3 = ctypes.c_float * 3
    _Skyrmion(ctypes.c_void_p(p_state), ctypes.c_float(radius), ctypes.c_float(order),
              ctypes.c_float(phase), ctypes.c_bool(up_down), ctypes.c_bool(achiral),
              ctypes.c_bool(right_left), vec3(*pos), vec3(*border_rectangular),
              ctypes.c_float(border_cylindrical), ctypes.c_float(border_spherical),
              ctypes.c_bool(inverted), ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

_Hopfion             = _spirit.Configuration_Hopfion
_Hopfion.argtypes    = [ctypes.c_void_p, ctypes.c_float, ctypes.c_int, ctypes.POINTER(ctypes.c_float),
                        ctypes.POINTER(ctypes.c_float), ctypes.c_float, ctypes.c_float, ctypes.c_bool,
                        ctypes.c_int, ctypes.c_int]
_Hopfion.restype     = None
def hopfion(p_state, radius, order=1, pos=[0,0,0], border_rectangular=[-1,-1,-1],
            border_cylindrical=-1, border_spherical=-1, inverted=False, idx_image=-1, idx_chain=-1):
    """Set a Hopfion configuration.

    - radius: the distance from the center to the center of the corresponding tubular isosurface
    - order: TODO

    In contrast to the skyrmion, it extends over the whole allowed space.
    """
    vec3 = ctypes.c_float * 3
    _Hopfion(ctypes.c_void_p(p_state), ctypes.c_float(radius), ctypes.c_int(order),
             vec3(*pos), vec3(*border_rectangular), ctypes.c_float(border_cylindrical),
             ctypes.c_float(border_spherical), ctypes.c_bool(inverted),
             ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

_SpinSpiral             = _spirit.Configuration_SpinSpiral
_SpinSpiral.argtypes    = [ctypes.c_void_p, ctypes.c_char_p, ctypes.POINTER(ctypes.c_float),
                           ctypes.POINTER(ctypes.c_float), ctypes.c_float,
                           ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float),
                           ctypes.c_float, ctypes.c_float, ctypes.c_bool,
                           ctypes.c_int, ctypes.c_int]
_SpinSpiral.restype     = None
def spin_spiral(p_state, direction_type, q_vector, axis, theta, pos=[0,0,0],
               border_rectangular=[-1,-1,-1], border_cylindrical=-1, border_spherical=-1,
               inverted=False, idx_image=-1, idx_chain=-1):
    """Set a spin spiral configuration.

    TODO: document parameters
    - direction_type:
    - q_vector:
    - axis:
    - theta:
    """
    vec3 = ctypes.c_float * 3
    _SpinSpiral(ctypes.c_void_p(p_state), ctypes.c_char_p(direction_type.encode('utf-8')),
                vec3(*q_vector), vec3(*axis), ctypes.c_float(theta), vec3(*pos),
                vec3(*border_rectangular), ctypes.c_float(border_cylindrical),
                ctypes.c_float(border_spherical), ctypes.c_bool(inverted),
                ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

_Set_Pinned             = _spirit.Configuration_Set_Pinned
_Set_Pinned.argtypes    = [ctypes.c_void_p, ctypes.c_bool, ctypes.POINTER(ctypes.c_float),
                    ctypes.POINTER(ctypes.c_float), ctypes.c_float, ctypes.c_float, ctypes.c_bool,
                    ctypes.c_int, ctypes.c_int]
_Set_Pinned.restype     = None
def set_pinned(p_state, pinned, pos=[0.0,0.0,0.0], border_rectangular=[-1.0,-1.0,-1.0], border_cylindrical=-1.0,
          border_spherical=-1.0, inverted=False, idx_image=-1, idx_chain=-1):
    """Set whether the spins within the given region are pinned or not.

    If they are pinned, they are pinned to their current orientation.
    """
    vec3 = ctypes.c_float * 3
    _Set_Pinned(ctypes.c_void_p(p_state), ctypes.c_bool(pinned), vec3(*pos), vec3(*border_rectangular),
           ctypes.c_float(border_cylindrical), ctypes.c_float(border_spherical),
           ctypes.c_bool(inverted), ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

_Set_Atom_Type             = _spirit.Configuration_Set_Atom_Type
_Set_Atom_Type.argtypes    = [ctypes.c_void_p, ctypes.c_int, ctypes.POINTER(ctypes.c_float),
                            ctypes.POINTER(ctypes.c_float), ctypes.c_float, ctypes.c_float, ctypes.c_bool,
                            ctypes.c_int, ctypes.c_int]
_Set_Atom_Type.restype     = None
def set_atom_type(p_state, atom_type=0, pos=[0.0,0.0,0.0], border_rectangular=[-1.0,-1.0,-1.0], border_cylindrical=-1.0,
          border_spherical=-1.0, inverted=False, idx_image=-1, idx_chain=-1):
    """Set the type of the atoms in the given region (default: 0).

    This can be used e.g. to insert defects (-1).
    """
    vec3 = ctypes.c_float * 3
    _Set_Atom_Type(ctypes.c_void_p(p_state), ctypes.c_int(atom_type), vec3(*pos), vec3(*border_rectangular),
           ctypes.c_float(border_cylindrical), ctypes.c_float(border_spherical),
           ctypes.c_bool(inverted), ctypes.c_int(idx_image), ctypes.c_int(idx_chain))


"""
Hamiltonian
====================

Set the parameters of the Heisenberg Hamiltonian, such as external field or exchange interaction.
"""

import spirit.spiritlib as spiritlib
import ctypes

### Load Library
_spirit = spiritlib.load_spirit_library()

### DM vector chirality
CHIRALITY_BLOCH         =  1
"""DMI Bloch chirality type for neighbour shells"""

CHIRALITY_NEEL          =  2
"""DMI Neel chirality type for neighbour shells"""

CHIRALITY_BLOCH_INVERSE = -1
"""DMI Bloch chirality type for neighbour shells with opposite sign"""

CHIRALITY_NEEL_INVERSE  = -2
"""DMI Neel chirality type for neighbour shells with opposite sign"""

### DDI METHOD
DDI_METHOD_NONE         = 0
"""Dipole-dipole interaction: do not calculate"""

DDI_METHOD_FFT          = 1
"""Dipole-dipole interaction: use FFT convolutions"""

DDI_METHOD_FMM          = 2
"""Dipole-dipole interaction: use a fast multipole method (FMM)"""

DDI_METHOD_CUTOFF       = 3
"""Dipole-dipole interaction: use a direct summation with a cutoff radius"""

### ---------------------------------- Set ----------------------------------

_Set_Boundary_Conditions             = _spirit.Hamiltonian_Set_Boundary_Conditions
_Set_Boundary_Conditions.argtypes    = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_bool),
                          ctypes.c_int, ctypes.c_int]
_Set_Boundary_Conditions.restype     = None
def set_boundary_conditions(p_state, boundaries, idx_image=-1, idx_chain=-1):
    """Set the boundary conditions along the translation directions [a, b, c].

    0 = open, 1 = periodical
    """
    bool3 = ctypes.c_bool * 3
    _Set_Boundary_Conditions(ctypes.c_void_p(p_state), bool3(*boundaries),
                             ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

_Set_Field             = _spirit.Hamiltonian_Set_Field
_Set_Field.argtypes    = [ctypes.c_void_p, ctypes.c_float, ctypes.POINTER(ctypes.c_float),
                          ctypes.c_int, ctypes.c_int]
_Set_Field.restype     = None
def set_field(p_state, magnitude, direction, idx_image=-1, idx_chain=-1):
    """Set the (homogeneous) external magnetic field."""
    vec3 = ctypes.c_float * 3
    _Set_Field(ctypes.c_void_p(p_state), ctypes.c_float(magnitude), vec3(*direction),
               ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

_Set_Anisotropy             = _spirit.Hamiltonian_Set_Anisotropy
_Set_Anisotropy.argtypes    = [ctypes.c_void_p, ctypes.c_float, ctypes.POINTER(ctypes.c_float),
                               ctypes.c_int, ctypes.c_int]
_Set_Anisotropy.restype     = None
def set_anisotropy(p_state, magnitude, direction, idx_image=-1, idx_chain=-1):
    """Set the (homogeneous) magnetocrystalline anisotropy."""
    vec3 = ctypes.c_float * 3
    _Set_Anisotropy(ctypes.c_void_p(p_state), ctypes.c_float(magnitude), vec3(*direction),
                    ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

_Set_Exchange             = _spirit.Hamiltonian_Set_Exchange
_Set_Exchange.argtypes    = [ctypes.c_void_p, ctypes.c_int, ctypes.POINTER(ctypes.c_float),
                               ctypes.c_int, ctypes.c_int]
_Set_Exchange.restype     = None
def set_exchange(p_state, n_shells, J_ij, idx_image=-1, idx_chain=-1):
    """Set the Exchange interaction in terms of neighbour shells."""
    vec = ctypes.c_float * n_shells
    _Set_Exchange(ctypes.c_void_p(p_state), ctypes.c_int(n_shells), vec(*J_ij),
                  ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

_Set_DMI             = _spirit.Hamiltonian_Set_DMI
_Set_DMI.argtypes    = [ctypes.c_void_p, ctypes.c_int, ctypes.POINTER(ctypes.c_float),
                        ctypes.c_int, ctypes.c_int, ctypes.c_int]
_Set_DMI.restype     = None
def set_dmi(p_state, n_shells, D_ij, chirality=CHIRALITY_BLOCH, idx_image=-1, idx_chain=-1):
    """Set the Dzyaloshinskii-Moriya interaction in terms of neighbour shells."""
    vec = ctypes.c_float * n_shells
    _Set_DMI(ctypes.c_void_p(p_state), ctypes.c_int(n_shells), vec(*D_ij),
             ctypes.c_int(chirality), ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

_Set_DDI             = _spirit.Hamiltonian_Set_DDI
_Set_DDI.argtypes    = [ctypes.c_void_p, ctypes.c_int, ctypes.POINTER(ctypes.c_int), ctypes.c_float,
                        ctypes.c_int, ctypes.c_int]
_Set_DDI.restype     = None
def set_ddi(p_state, ddi_method, n_periodic_images=[4,4,4], radius=0.0, idx_image=-1, idx_chain=-1):
    """Set the dipolar interaction calculation method.

    - ddi_method -- one of the integers defined above
    - n_periodic_images -- the number of periodical images in the three translation directions, taken into account
      when boundaries in the corresponding direction are periodical
    - radius -- the cutoff radius for the direct summation method
    """
    vec3 = ctypes.c_int * 3
    _Set_DDI(ctypes.c_void_p(p_state), ctypes.c_int(ddi_method) , vec3(*n_periodic_images), ctypes.c_float(radius),
             ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

### ---------------------------------- Get ----------------------------------

_Get_Name          = _spirit.Hamiltonian_Get_Name
_Get_Name.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Get_Name.restype  = ctypes.c_char_p
def get_name(p_state, idx_image=-1, idx_chain=-1):
    """Returns a string containing the name of the Hamiltonian currently in use."""
    return str(_Get_Name(ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)))

_Get_Boundary_Conditions          = _spirit.Hamiltonian_Get_Boundary_Conditions
_Get_Boundary_Conditions.argtypes = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_bool),
                                     ctypes.c_int, ctypes.c_int]
_Get_Boundary_Conditions.restype  = None
def get_boundary_conditions(p_state, idx_image=-1, idx_chain=-1):
    """Returns an array of shape (3) containing the boundary conditions in the
    three translation directions [a, b, c] of the lattice.
    """
    boundaries = (3*ctypes.c_bool)()
    _Get_Boundary_Conditions(ctypes.c_void_p(p_state), boundaries,
                             ctypes.c_int(idx_image), ctypes.c_int(idx_chain))
    return [bc for bc in boundaries]

_Get_Field          = _spirit.Hamiltonian_Get_Field
_Get_Field.argtypes = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_float),
                       ctypes.POINTER(ctypes.c_float), ctypes.c_int, ctypes.c_int]
_Get_Field.restype  = None
def get_field(p_state, idx_image=-1, idx_chain=-1):
    """Returns the magnitude and an array of shape (3) containing the direction of
    the external magnetic field.
    """
    magnitude = (1*ctypes.c_float)()
    normal = (3*ctypes.c_float)()
    _Get_Field(ctypes.c_void_p(p_state), magnitude, normal,
               ctypes.c_int(idx_image), ctypes.c_int(idx_chain))
    return float(magnitude), [n for n in normal]

_Get_DDI          = _spirit.Hamiltonian_Get_DDI
_Get_DDI.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Get_DDI.restype  = ctypes.c_float
def get_ddi(p_state, idx_image=-1, idx_chain=-1):
    """Returns the cutoff radius of the DDI."""
    return float(_Get_DDI(ctypes.c_void_p(p_state), ctypes.c_int(idx_image),
                          ctypes.c_int(idx_chain)))


"""
Geodesic nudged elastic band (GNEB)
-------------------------------------------------------------
"""

import spirit.spiritlib as spiritlib
from spirit.io import FILEFORMAT_OVF_TEXT
import ctypes

### Load Library
_spirit = spiritlib.load_spirit_library()

# GNEB image types
IMAGE_NORMAL     = 0
"""Regular GNEB image type."""

IMAGE_CLIMBING   = 1
"""Climbing GNEB image type.
Climbing images move towards maxima along the path.
"""

IMAGE_FALLING    = 2
"""Falling GNEB image type.
Falling images move towards the closest minima.
"""

IMAGE_STATIONARY = 3
"""Stationary GNEB image type.
Stationary images are not influenced during a GNEB calculation.
"""

### ---------------------------------- Set ----------------------------------

_GNEB_Set_Output_Tag          = _spirit.Parameters_GNEB_Set_Output_Tag
_GNEB_Set_Output_Tag.argtypes = [ctypes.c_void_p, ctypes.c_char_p,
                                ctypes.c_int, ctypes.c_int]
_GNEB_Set_Output_Tag.restype  = None
def set_output_tag(p_state, tag, idx_image=-1, idx_chain=-1):
    """Set the tag placed in front of output file names.

    If the tag is "<time>", it will be the date-time of the creation of the state."""
    _GNEB_Set_Output_Tag(ctypes.c_void_p(p_state), ctypes.c_char_p(tag.encode('utf-8')),
                        ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

_GNEB_Set_Output_Folder          = _spirit.Parameters_GNEB_Set_Output_Folder
_GNEB_Set_Output_Folder.argtypes = [ctypes.c_void_p, ctypes.c_char_p,
                                    ctypes.c_int, ctypes.c_int]
_GNEB_Set_Output_Folder.restype  = None
def set_output_folder(p_state, folder, idx_image=-1, idx_chain=-1):
    """Set the folder, where output files are placed."""
    _GNEB_Set_Output_Folder(ctypes.c_void_p(p_state), ctypes.c_char_p(folder.encode('utf-8')),
                        ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

_GNEB_Set_Output_General          = _spirit.Parameters_GNEB_Set_Output_General
_GNEB_Set_Output_General.argtypes = [ctypes.c_void_p, ctypes.c_bool, ctypes.c_bool,
                                    ctypes.c_bool, ctypes.c_int, ctypes.c_int]
_GNEB_Set_Output_General.restype  = None
def set_output_general(p_state, any=True, initial=False, final=False, idx_image=-1, idx_chain=-1):
    """Set whether to write any output files at all."""
    _GNEB_Set_Output_General(ctypes.c_void_p(p_state), ctypes.c_bool(any), ctypes.c_bool(initial),
                        ctypes.c_bool(final), ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

_GNEB_Set_Output_Energies          = _spirit.Parameters_GNEB_Set_Output_Energies
_GNEB_Set_Output_Energies.argtypes = [ctypes.c_void_p, ctypes.c_bool, ctypes.c_bool,
                                    ctypes.c_bool, ctypes.c_bool, ctypes.c_int, ctypes.c_int]
_GNEB_Set_Output_Energies.restype  = None
def set_output_energies(p_state, step=True, interpolated=True, divide_by_nos=True, add_readability_lines=True, idx_image=-1, idx_chain=-1):
    """Set whether to write energy output files.

    - step: whether to write a new file after each set of iterations
    - interpolated: whether to write a file containing interpolated reaction coordinate and energy values
    - divide_by_nos: whether to divide energies by the number of spins
    - add_readability_lines: whether to separate columns by lines
    """
    _GNEB_Set_Output_Energies(ctypes.c_void_p(p_state), ctypes.c_bool(step), ctypes.c_bool(interpolated),
                        ctypes.c_bool(divide_by_nos), ctypes.c_bool(add_readability_lines),
                        ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

_GNEB_Set_Output_Chain          = _spirit.Parameters_GNEB_Set_Output_Chain
_GNEB_Set_Output_Chain.argtypes = [ctypes.c_void_p, ctypes.c_bool, ctypes.c_int, ctypes.c_int, ctypes.c_int]
_GNEB_Set_Output_Chain.restype  = None
def set_output_chain(p_state, step=False, filetype=FILEFORMAT_OVF_TEXT, idx_image=-1, idx_chain=-1):
    """Set whether to write chain output files.

    - step: whether to write a new file after each set of iterations
    - filetype: the format in which the data is written
    """
    _GNEB_Set_Output_Chain(ctypes.c_void_p(p_state), ctypes.c_bool(step), ctypes.c_int(filetype),
                        ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

_GNEB_Set_N_Iterations             = _spirit.Parameters_GNEB_Set_N_Iterations
_GNEB_Set_N_Iterations.argtypes    = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int,
                                      ctypes.c_int, ctypes.c_int]
_GNEB_Set_N_Iterations.restype     = None
def set_iterations(p_state, n_iterations, n_iterations_log, idx_image=-1, idx_chain=-1):
    """Set the number of iterations and how often to log and write output.

    - n_iterations: the maximum number of iterations
    - n_iterations_log: the number of iterations after which status is logged and output written
    """
    _GNEB_Set_N_Iterations(ctypes.c_void_p(p_state), ctypes.c_int(n_iterations),
                           ctypes.c_int(n_iterations_log), ctypes.c_int(idx_image),
                           ctypes.c_int(idx_chain))

_GNEB_Set_Convergence           = _spirit.Parameters_GNEB_Set_Convergence
_GNEB_Set_Convergence.argtypes  = [ctypes.c_void_p, ctypes.c_float, ctypes.c_int, ctypes.c_int]
_GNEB_Set_Convergence.restype   = None
def set_convergence(p_state, convergence, idx_image=-1, idx_chain=-1):
    """Set the convergence limit.

    When the maximum absolute component value of the force drops below this value,
    the calculation is considered converged and will stop.
    """
    _GNEB_Set_Convergence(ctypes.c_void_p(p_state), ctypes.c_float(convergence),
                          ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

_GNEB_Set_Spring_Constant          = _spirit.Parameters_GNEB_Set_Spring_Constant
_GNEB_Set_Spring_Constant.argtypes = [ctypes.c_void_p, ctypes.c_float, ctypes.c_int, ctypes.c_int]
_GNEB_Set_Spring_Constant.restype  = None
_GNEB_Set_Spring_Force_Ratio          = _spirit.Parameters_GNEB_Set_Spring_Force_Ratio
_GNEB_Set_Spring_Force_Ratio.argtypes = [ctypes.c_void_p, ctypes.c_float, ctypes.c_int, ctypes.c_int]
_GNEB_Set_Spring_Force_Ratio.restype  = None
def set_spring_force(p_state, spring_constant=1, ratio=0, idx_image=-1, idx_chain=-1):
    """Set the spring force constant and the ratio between energy and reaction coordinate."""
    _GNEB_Set_Spring_Constant(ctypes.c_void_p(p_state), ctypes.c_float(spring_constant),
                              ctypes.c_int(idx_image), ctypes.c_int(idx_chain))
    _GNEB_Set_Spring_Force_Ratio(ctypes.c_void_p(p_state), ctypes.c_float(ratio),
                              ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

_GNEB_Set_Path_Shortening_Constant           = _spirit.Parameters_GNEB_Set_Path_Shortening_Constant
_GNEB_Set_Path_Shortening_Constant.argtypes  = [ctypes.c_void_p, ctypes.c_float, ctypes.c_int, ctypes.c_int]
_GNEB_Set_Path_Shortening_Constant.restype   = None
def set_path_shortening_constant(p_state, shortening_constant, idx_image=-1, idx_chain=-1):
    """Set the path shortening constant."""
    _GNEB_Set_Path_Shortening_Constant(ctypes.c_void_p(p_state), ctypes.c_float(shortening_constant),
                          ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

_GNEB_Set_Climbing_Falling             = _spirit.Parameters_GNEB_Set_Climbing_Falling
_GNEB_Set_Climbing_Falling.argtypes    = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int, ctypes.c_int]
_GNEB_Set_Climbing_Falling.restype     = None
def set_climbing_falling(p_state, image_type, idx_image=-1, idx_chain=-1):
    """Set the GNEB image type (see the integers defined above)."""
    _GNEB_Set_Climbing_Falling(ctypes.c_void_p(p_state), ctypes.c_int(image_type),
                               ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

_GNEB_Set_Image_Type_Automatically             = _spirit.Parameters_GNEB_Set_Image_Type_Automatically
_GNEB_Set_Image_Type_Automatically.argtypes    = [ctypes.c_void_p, ctypes.c_int]
_GNEB_Set_Image_Type_Automatically.restype     = None
def set_image_type_automatically(p_state, idx_chain=-1):
    """Automatically set GNEB image types.

    Minima along the path will be set to falling, maxima to climbing and the rest to regular.
    """
    _GNEB_Set_Image_Type_Automatically(ctypes.c_void_p(p_state), ctypes.c_int(idx_chain))

### ---------------------------------- Get ----------------------------------

_GNEB_Get_N_Iterations             = _spirit.Parameters_GNEB_Get_N_Iterations
_GNEB_Get_N_Iterations.argtypes    = [ctypes.c_void_p, ctypes.POINTER( ctypes.c_int ),
                                      ctypes.POINTER( ctypes.c_int ), ctypes.c_int]
_GNEB_Get_N_Iterations.restype     = None
def get_iterations(p_state, idx_image=-1, idx_chain=-1):
    """Returns the maximum number of iterations and the step size."""
    n_iterations = ctypes.c_int()
    n_iterations_log = ctypes.c_int()
    _GNEB_Get_N_Iterations(ctypes.c_void_p(p_state), ctypes.pointer(n_iterations),
                           ctypes.pointer(n_iterations_log), ctypes.c_int(idx_image),
                           ctypes.c_int(idx_chain) )
    return int(n_iterations.value), int(n_iterations_log.value)

_GNEB_Get_Convergence           = _spirit.Parameters_GNEB_Get_Convergence
_GNEB_Get_Convergence.argtypes  = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_GNEB_Get_Convergence.restype   = ctypes.c_float
def get_convergence(p_state, idx_image=-1, idx_chain=-1):
    """Returns the convergence value."""
    return float( _GNEB_Get_Convergence(ctypes.c_void_p(p_state),
                                        ctypes.c_int(idx_image), ctypes.c_int(idx_chain)))

_GNEB_Get_Spring_Constant          = _spirit.Parameters_GNEB_Get_Spring_Constant
_GNEB_Get_Spring_Constant.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_GNEB_Get_Spring_Constant.restype  = ctypes.c_float
_GNEB_Get_Spring_Force_Ratio          = _spirit.Parameters_GNEB_Get_Spring_Force_Ratio
_GNEB_Get_Spring_Force_Ratio.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_GNEB_Get_Spring_Force_Ratio.restype  = ctypes.c_float
def get_spring_force(p_state, idx_image=-1, idx_chain=-1):
    """Returns the spring force constant and Ratio of energy to reaction coordinate."""
    constant = float(_GNEB_Get_Spring_Constant(ctypes.c_void_p(p_state),
                                           ctypes.c_int(idx_image), ctypes.c_int(idx_chain)))
    ratio    = float(_GNEB_Get_Spring_Force_Ratio(ctypes.c_void_p(p_state),
                                           ctypes.c_int(idx_image), ctypes.c_int(idx_chain)))
    return constant, ratio

_GNEB_Get_Path_Shortening_Constant           = _spirit.Parameters_GNEB_Get_Path_Shortening_Constant
_GNEB_Get_Path_Shortening_Constant.argtypes  = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_GNEB_Get_Path_Shortening_Constant.restype   = ctypes.c_float
def get_path_shortening_constant(p_state, idx_image=-1, idx_chain=-1):
    """Return the path shortening constant."""
    return float( _GNEB_Get_Path_Shortening_Constant(ctypes.c_void_p(p_state),
                                        ctypes.c_int(idx_image), ctypes.c_int(idx_chain)))

_GNEB_Get_Climbing_Falling             = _spirit.Parameters_GNEB_Get_Climbing_Falling
_GNEB_Get_Climbing_Falling.argtypes    = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_GNEB_Get_Climbing_Falling.restype     = ctypes.c_int
def get_climbing_falling(p_state, idx_image=-1, idx_chain=-1):
    """Returns the integer of whether an image is regular, climbing, falling, or stationary.

    The integers are defined above."""
    return int(_GNEB_Get_Climbing_Falling(ctypes.c_void_p(p_state),
                                          ctypes.c_int(idx_image), ctypes.c_int(idx_chain)))

_GNEB_Get_N_Energy_Interpolations             = _spirit.Parameters_GNEB_Get_N_Energy_Interpolations
_GNEB_Get_N_Energy_Interpolations.argtypes    = [ctypes.c_void_p, ctypes.c_int]
_GNEB_Get_N_Energy_Interpolations.restype     = ctypes.c_int
def get_n_energy_interpolations(p_state, idx_chain=-1):
    """Returns the number of energy values interpolated between images."""
    return int(_GNEB_Get_N_Energy_Interpolations(ctypes.c_void_p(p_state), ctypes.c_int(idx_chain)))


"""
System
====================
"""

import spirit.spiritlib as spiritlib
import ctypes

### Load Library
_spirit = spiritlib.load_spirit_library()

from spirit.scalar import scalar
from spirit import parameters
from numpy import frombuffer, ndarray as np

### Get Chain index
_Get_Index          = _spirit.System_Get_Index
_Get_Index.argtypes = [ctypes.c_void_p]
_Get_Index.restype  = ctypes.c_int
def get_index(p_state):
    """Returns the index of the currently active image."""
    return int(_Get_Index(ctypes.c_void_p(p_state)))

### Get Chain number of images
_Get_NOS            = _spirit.System_Get_NOS
_Get_NOS.argtypes   = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Get_NOS.restype    = ctypes.c_int
def get_nos(p_state, idx_image=-1, idx_chain=-1):
    """Returns the number of spins (NOS)."""
    return int(_Get_NOS(ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)))

### Get Pointer to Spin Directions
# NOTE: Changing the values of the array_view one can alter the value of the data of the state
_Get_Spin_Directions            = _spirit.System_Get_Spin_Directions
_Get_Spin_Directions.argtypes   = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Get_Spin_Directions.restype    = ctypes.POINTER(scalar)
def get_spin_directions(p_state, idx_image=-1, idx_chain=-1):
    """Returns an `numpy.array_view` of shape (NOS, 3) with the components of each spins orientation vector.

    Changing the contents of this array_view will have direct effect on calculations etc.
    """
    nos = get_nos(p_state, idx_image, idx_chain)
    ArrayType = scalar*3*nos
    Data = _Get_Spin_Directions(ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain))
    array_pointer = ctypes.cast(Data, ctypes.POINTER(ArrayType))
    array = frombuffer(array_pointer.contents, dtype=scalar)
    array_view = array.view()
    array_view.shape = (nos, 3)
    return array_view

### Get Pointer to Effective Field
# NOTE: Changing the values of the array_view one can alter the value of the data of the state
_Get_Effective_Field            = _spirit.System_Get_Effective_Field
_Get_Effective_Field.argtypes   = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Get_Effective_Field.restype    = ctypes.POINTER(scalar)
def get_effective_field(p_state, idx_image=-1, idx_chain=-1):
    nos = get_nos(p_state, idx_image, idx_chain)
    ArrayType = scalar*3*nos
    Data = _Get_Effective_Field(ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain))
    array_pointer = ctypes.cast(Data, ctypes.POINTER(ArrayType))
    array = frombuffer(array_pointer.contents, dtype=scalar)
    array_view = array.view()
    array_view.shape = (nos, 3)
    return array_view

### Get Pointer to an eigenmode
# NOTE: Changing the values of the array_view one can alter the value of the data of the state
_Get_Eigenmode            = _spirit.System_Get_Eigenmode
_Get_Eigenmode.argtypes   = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int, ctypes.c_int]
_Get_Eigenmode.restype    = ctypes.POINTER(scalar)
def get_eigenmode(p_state, idx_mode, idx_image=-1, idx_chain=-1):
    nos = get_nos(p_state, idx_image, idx_chain)
    ArrayType = scalar*3*nos
    Data = _Get_Eigenmode(ctypes.c_void_p(p_state), ctypes.c_int(idx_mode),
                          ctypes.c_int(idx_image), ctypes.c_int(idx_chain))
    array_pointer = ctypes.cast(Data, ctypes.POINTER(ArrayType))
    array = frombuffer(array_pointer.contents, dtype=scalar)
    array_view = array.view()
    array_view.shape = (nos, 3)
    return array_view

### Get total Energy
_Get_Energy          = _spirit.System_Get_Energy
_Get_Energy.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Get_Energy.restype  = ctypes.c_float
def get_energy(p_state, idx_image=-1, idx_chain=-1):
    """Calculates and returns the energy of the system."""
    return float(_Get_Energy(ctypes.c_void_p(p_state), ctypes.c_int(idx_image),
                             ctypes.c_int(idx_chain)))

### Get Energy
_Get_Eigenvalues          = _spirit.System_Get_Eigenvalues
_Get_Eigenvalues.argtypes = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_float), ctypes.c_int, ctypes.c_int]
_Get_Eigenvalues.restype  = None
def get_eigenvalues(p_state, idx_image=-1, idx_chain=-1):
    noe = parameters.ema.getNModes(p_state, idx_image, idx_chain)
    eigenvalues = (noe*ctypes.c_float)()
    _Get_Eigenvalues(ctypes.c_void_p(p_state), eigenvalues, ctypes.c_int(idx_image), ctypes.c_int(idx_chain))
    return eigenvalues

# NOTE: excluded since there is no clean way to get the C++ pairs
### Get Energy array
# _Get_Energy_Array          = _spirit.System_Get_Energy_Array
# _Get_Energy_Array.argtypes = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_float),
#                               ctypes.c_int, ctypes.c_int]
# _Get_Energy_Array.restype  = None
# def Get_Energy_Array(p_state, idx_image=-1, idx_chain=-1):
#     Energies
#     _Get_Energy_Array(ctypes.c_void_p(p_state), energies,
#                       ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

### Get Chain number of images
_Update_Data            = _spirit.System_Update_Data
_Update_Data.argtypes   = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Update_Data.restype    = None
def update_data(p_state, idx_image=-1, idx_chain=-1):
    """TODO: document when this needs to be called."""
    _Update_Data(ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain))


### Eigenmodes
_Eigenmodes          = _spirit.System_Update_Eigenmodes
_Eigenmodes.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Eigenmodes.restype  = None
def update_eigenmodes(p_state, idx_image=-1, idx_chain=-1):
    """Calculates eigenmodes of a system according to EMA parameters.
    This needs to be called or eigenmodes need to be read in before they can be used by other functions
    (e.g. writing them to a file).
    """
    spiritlib.wrap_function(_Eigenmodes, [ctypes.c_void_p(p_state), ctypes.c_int(idx_image),
                                         ctypes.c_int(idx_chain)])

### Print Energy array
_Print_Energy_Array            = _spirit.System_Print_Energy_Array
_Print_Energy_Array.argtypes   = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Print_Energy_Array.restype    = None
def print_energy_array(p_state, idx_image=-1, idx_chain=-1):
    """Print the energy array of the state to the console."""
    _Print_Energy_Array(ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain))


"""
Log
====================
"""

import spirit.spiritlib as spiritlib
import ctypes

### Load Library
_spirit = spiritlib.load_spirit_library()

# Log levels
LEVEL_ALL       = 0
LEVEL_SEVERE    = 1
LEVEL_ERROR     = 2
LEVEL_WARNING   = 3
LEVEL_PARAMETER = 4
LEVEL_INFO      = 5
LEVEL_DEBUG     = 6

# Log message senders
SENDER_ALL  = 0
SENDER_IO   = 1
SENDER_GNEB = 2
SENDER_LLG  = 3
SENDER_MC   = 4
SENDER_MMF  = 5
SENDER_API  = 6
SENDER_UI   = 7

_Send          = _spirit.Log_Send
_Send.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int, ctypes.c_char_p,
                  ctypes.c_int, ctypes.c_int]
_Send.restype  = None
def send(p_state, level, sender, message, idx_image=-1, idx_chain=-1):
    """Add a message to the log.

    - level: see integers defined above. The message may be printed to the console and/or written
      to the log file, depending on the current log parameters
    - sender: see integers defined above. Used to distinguish context
    - message: a string which to log
    - idx_image: can be used to specify to which image the message relates (default: active image)
    """
    _Send(ctypes.c_void_p(p_state), ctypes.c_int(level), ctypes.c_int(sender),
          ctypes.c_char_p(message.encode('utf-8')), ctypes.c_int(idx_image), ctypes.c_int(idx_chain))

_Append          = _spirit.Log_Append
_Append.argtypes = [ctypes.c_void_p]
_Append.restype  = None
def append(p_state):
    """Force the appending of new messages to the log file."""
    _Append(ctypes.c_void_p(p_state))

_Get_N_Entries          = _spirit.Log_Get_N_Entries
_Get_N_Entries.argtypes = [ctypes.c_void_p]
_Get_N_Entries.restype  = ctypes.c_int
def get_n_entries(p_state):
    """Returns the number of Log entries."""
    return int(_Get_N_Entries(ctypes.c_void_p(p_state)))

_Get_N_Errors          = _spirit.Log_Get_N_Errors
_Get_N_Errors.argtypes = [ctypes.c_void_p]
_Get_N_Errors.restype  = ctypes.c_int
def get_n_errors(p_state):
    """Returns the number of error messages that have been logged."""
    return int(_Get_N_Errors(ctypes.c_void_p(p_state)))

_Get_N_Warnings          = _spirit.Log_Get_N_Warnings
_Get_N_Warnings.argtypes = [ctypes.c_void_p]
_Get_N_Warnings.restype  = ctypes.c_int
def get_n_warnings(p_state):
    """Returns the number of warning messages that have been logged."""
    return int(_Get_N_Warnings(ctypes.c_void_p(p_state)))

_Set_Output_File_Tag          = _spirit.Log_Set_Output_File_Tag
_Set_Output_File_Tag.argtypes = [ctypes.c_void_p, ctypes.c_char_p]
_Set_Output_File_Tag.restype  = None
def set_output_file_tag(p_state, tag):
    """Set the tagging string which is placed in front of the log file.

    If "<time>" is used for the tag, it will be the time of creation of the state.
    """
    _Set_Output_File_Tag(ctypes.c_void_p(p_state), ctypes.c_char_p(tag.encode('utf-8')))

_Set_Output_Folder          = _spirit.Log_Set_Output_Folder
_Set_Output_Folder.argtypes = [ctypes.c_void_p, ctypes.c_char_p]
_Set_Output_Folder.restype  = None
def set_output_folder(p_state, tag):
    """Set the output folder in which to place the log file."""
    _Set_Output_Folder(ctypes.c_void_p(p_state), ctypes.c_char_p(tag.encode('utf-8')))

_Set_Output_To_Console          = _spirit.Log_Set_Output_To_Console
_Set_Output_To_Console.argtypes = [ctypes.c_void_p, ctypes.c_bool, ctypes.c_int]
_Set_Output_To_Console.restype  = None
def set_output_to_console(p_state, output, level):
    """Set whether the Log is output to the console and the level up to which messages are logged."""
    _Set_Output_To_Console(ctypes.c_void_p(p_state), ctypes.c_bool(output), ctypes.c_int(level))

_Set_Output_To_File          = _spirit.Log_Set_Output_To_File
_Set_Output_To_File.argtypes = [ctypes.c_void_p, ctypes.c_bool, ctypes.c_int]
_Set_Output_To_File.restype  = None
def set_outputToFile(p_state, output, level):
    """Set whether the Log is output to a file and the level up to which messages are logged."""
    _Set_Output_To_File(ctypes.c_void_p(p_state), ctypes.c_bool(output), ctypes.c_int(level))

_Get_Output_To_Console          = _spirit.Log_Get_Output_To_Console
_Get_Output_To_Console.argtypes = [ctypes.c_void_p]
_Get_Output_To_Console.restype  = ctypes.c_bool
def GetOutputToConsole(p_state):
    """Returns a bool indicating whether the Log is output to the console."""
    return bool(_Get_Output_To_Console(ctypes.c_void_p(p_state)))

_Get_Output_Console_Level          = _spirit.Log_Get_Output_Console_Level
_Get_Output_Console_Level.argtypes = [ctypes.c_void_p]
_Get_Output_Console_Level.restype  = ctypes.c_int
def GetOutputConsoleLevel(p_state):
    """Returns the level up to which the Log is output to the console.

    The return value will be one of the integers defined above.
    """
    return int(_Get_Output_Console_Level(ctypes.c_void_p(p_state)))

_Get_Output_To_File          = _spirit.Log_Get_Output_To_File
_Get_Output_To_File.argtypes = [ctypes.c_void_p]
_Get_Output_To_File.restype  = ctypes.c_bool
def GetOutputToFile(p_state):
    """Returns a bool indicating whether the Log is output to a file."""
    return bool(_Get_Output_To_File(ctypes.c_void_p(p_state)))

_Get_Output_File_Level          = _spirit.Log_Get_Output_File_Level
_Get_Output_File_Level.argtypes = [ctypes.c_void_p]
_Get_Output_File_Level.restype  = ctypes.c_int
def GetOutputFileLevel(p_state):
    """Returns the level up to which the Log is output to a file.

    The return value will be one of the integers defined above.
    """
    return int(_Get_Output_File_Level(ctypes.c_void_p(p_state)))


"""
HTST
====================

Harmonic transition state theory.

Note that `calculate_prefactor` needs to be called before using any of the getter functions.
"""

import spirit.spiritlib as spiritlib
import spirit.parameters as parameters
import spirit.system as system
import ctypes

### Load Library
_spirit = spiritlib.load_spirit_library()


_Calculate          = _spirit.HTST_Calculate
_Calculate.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int, ctypes.c_int]
_Calculate.restype  = ctypes.c_float
def calculate(p_state, idx_image_minimum, idx_image_sp, idx_chain=-1):
    """Performs an HTST calculation and returns rate prefactor.

    *Note:* this function must be called before any of the getters.
    """
    return _Calculate(p_state, idx_image_minimum, idx_image_sp, idx_chain)


### Get HTST transition rate components
_Get_Info          = _spirit.HTST_Get_Info
_Get_Info.argtypes = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float),
                        ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float),
                        ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float),
                        ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.c_int]
_Get_Info.restype  = None
def get_info(p_state, idx_chain=-1):
    """Returns a set of HTST information:

    - the exponent of the temperature-dependence
    - `me`
    - `Omega_0`
    - `s`
    - zero mode volume at the minimum
    - zero mode volume at the saddle point
    - dynamical prefactor
    - full rate prefactor (without temperature dependent part)
    """
    temperature_exponent = ctypes.c_float()
    me                   = ctypes.c_float()
    Omega_0              = ctypes.c_float()
    s                    = ctypes.c_float()
    volume_min           = ctypes.c_float()
    volume_sp            = ctypes.c_float()
    prefactor_dynamical  = ctypes.c_float()
    prefactor            = ctypes.c_float()

    _Get_Info(ctypes.c_void_p(p_state), ctypes.pointer(temperature_exponent), ctypes.pointer(me),
                ctypes.pointer(Omega_0), ctypes.pointer(s), ctypes.pointer(volume_min),
                ctypes.pointer(volume_sp), ctypes.pointer(prefactor_dynamical),
                ctypes.pointer(prefactor), ctypes.c_int(idx_chain))

    return temperature_exponent.value, me.value, Omega_0.value, s.value,\
            volume_min.value, volume_sp.value, prefactor_dynamical.value, prefactor.value


_Get_Eigenvalues_Min          = _spirit.HTST_Get_Eigenvalues_Min
_Get_Eigenvalues_Min.argtypes = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_float), ctypes.c_int]
_Get_Eigenvalues_Min.restype  = None
def get_eigenvalues_min(p_state, idx_chain=-1):
    """Returns the eigenvalues at the minimum.
    Shape (2*nos)
    """
    nos             = system.get_nos(p_state, -1, idx_chain)
    eigenvalues_min = (2*nos*ctypes.c_float)()
    _Get_Eigenvalues_Min(ctypes.c_void_p(p_state), eigenvalues_min, ctypes.c_int(idx_chain))
    return eigenvalues_min


_Get_Eigenvectors_Min          = _spirit.HTST_Get_Eigenvectors_Min
_Get_Eigenvectors_Min.argtypes = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_float), ctypes.c_int]
_Get_Eigenvectors_Min.restype  = None
def get_eigenvectors_min(p_state, idx_chain=-1):
    """Returns the eigenvectors at the minimum.
    Shape (2*nos*nos)
    """
    nos              = system.get_nos(p_state, -1, idx_chain)
    eigenvectors_min = (2*nos*nos*ctypes.c_float)()
    _Get_Eigenvectors_Min(ctypes.c_void_p(p_state), eigenvectors_min, ctypes.c_int(idx_chain))
    return eigenvectors_min


_Get_Eigenvalues_SP          = _spirit.HTST_Get_Eigenvalues_SP
_Get_Eigenvalues_SP.argtypes = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_float), ctypes.c_int]
_Get_Eigenvalues_SP.restype  = None
def get_eigenvalues_sp(p_state, idx_chain=-1):
    """Returns the eigenvalues at the saddle point.
    Shape (2*nos)
    """
    nos            = system.get_nos(p_state, -1, idx_chain)
    eigenvalues_sp = (2*nos*ctypes.c_float)()
    _Get_Eigenvalues_SP(ctypes.c_void_p(p_state), eigenvalues_sp, ctypes.c_int(idx_chain))
    return eigenvalues_sp


_Get_Eigenvectors_SP          = _spirit.HTST_Get_Eigenvectors_SP
_Get_Eigenvectors_SP.argtypes = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_float), ctypes.c_int]
_Get_Eigenvectors_SP.restype  = None
def get_eigenvectors_sp(p_state, idx_chain=-1):
    """Returns the eigenvectors at the saddle point.
    Shape (2*nos*nos)
    """
    nos                 = system.get_nos(p_state, -1, idx_chain)
    eigenvectors_sp     = (2*nos*nos*ctypes.c_float)()
    _Get_Eigenvectors_SP(ctypes.c_void_p(p_state), eigenvectors_sp, ctypes.c_int(idx_chain))
    return eigenvectors_sp


_Get_Velocities          = _spirit.HTST_Get_Velocities
_Get_Velocities.argtypes = [ctypes.c_void_p, ctypes.POINTER(ctypes.c_float),
                            ctypes.POINTER(ctypes.c_float), ctypes.c_int]
_Get_Velocities.restype  = None
def get_velocities(p_state, idx_chain=-1):
    """Returns the velocities perpendicular to the dividing surface.
    Shape (2*nos)
    """
    nos        = system.get_nos(p_state, -1, idx_chain)
    velocities = (2*nos*ctypes.c_float)()
    _Get_Velocities(ctypes.c_void_p(p_state), velocities, ctypes.c_int(idx_chain))
    return velocities